#include <stdio.h>
#include <CL/cl.h>

void Explicit(cl_command_queue command_queue, size_t* global_item_size, size_t* local_item_size, cl_kernel kernel, cl_mem x0_mem_obj, cl_mem x_mem_obj, cl_mem r_mem_obj, double* x, double* r, int size_x, int size_y, double* error)
{
  int status;
  // local variables 
  int i,j;
  double result;

  // Process the entire lists
  if(clEnqueueNDRangeKernel(command_queue, kernel, 2, NULL, 
        global_item_size, local_item_size, 0, NULL, NULL) != CL_SUCCESS)
    fprintf(stderr,"Error in clEnqueueNDRangeKernel\n");

  // Read the buffer back to the array
  if(clEnqueueReadBuffer(command_queue, x_mem_obj, CL_TRUE, 0, 
        (size_x+2) * (size_y+2) * sizeof(double), x, 0, NULL, NULL) != CL_SUCCESS)
    fprintf(stderr,"Error in clEnqueueReadBuffer with x_mem_obj\n");

  if(clEnqueueReadBuffer(command_queue, r_mem_obj, CL_TRUE, 0, 
        (size_x+2) * (size_y+2) * sizeof(double), r, 0, NULL, NULL) != CL_SUCCESS)	    
    fprintf(stderr,"Error in clEnqueueReadBuffer with r_mem_obj\n");

  status = clFlush(command_queue);

  if(status)
  {fprintf(stderr,"Failed to flush command Queue\n");
    exit(-1);}

  if(clEnqueueWriteBuffer(command_queue, x0_mem_obj, CL_TRUE, 0, 
        (size_x+2) * (size_y+2) * sizeof(cl_double), x, 0, NULL, NULL) != CL_SUCCESS)	   	    
    fprintf(stderr,"Error in clEnqueueWriteuffer with x0_mem_obj\n");

  // Set new Argument - Outputs become Inputs
  status = clSetKernelArg(
      kernel,
      5,
      sizeof(cl_mem),
      (void*)&x0_mem_obj);

  if(status)
  {fprintf(stderr,"clSetKernelArg failed\n");
    exit(-1);}

  result = 0.0;
  for(i=1;i<=size_y;i++)
    for(j=1;j<=size_x;j++)
    {result = result + r[i*(size_x+2)+j];}

  *error=result;
} 	   

/* **************************************************************************
   !
   ! This subroutine setups the initial guess, i.e. the initial temperature
   ! within the domain
   !
   !************************************************************************** */

void initvalues(double* x0, int x_dim, int y_dim, double temp1_init, double temp2_init )
{
  /* local variables */
  int i,j;

  for(i=0;i<=x_dim+1;i++)
  {   
    x0[i] = temp1_init;
    x0[(x_dim+2)*(y_dim+1)+i] = temp1_init;
  }

  for(i=0;i<=y_dim+1;i++)
  {
    x0[i*(x_dim+2)] = temp1_init; 
    x0[i*(x_dim+2)+x_dim+1] = temp1_init;
  }     

  /* setup a constant initial state */

  for(j=1;j<=y_dim;j++)
  {
    for(i=1;i<=x_dim;i++)
    {
      x0[i+j*(x_dim+2)] = temp2_init;
    }
  }
}
