#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mpi.h"
#include "explicitMPI.h"

#define min(a,b) a <= b ? a : b

int main(int argc, char *argv[]) 
{
  /* size of the discretization */
  double** x;
  double* x_vals;
  double** x0;
  double* x0_vals;
  double* xfinal;
  double* xtemp;
  double dt, dt1, dt2, hx, hy, epsilon;
  double resLoc, result, t, tstart, tend, elapsed_time;

  int i,j,k,l;
  int convergence = 0;
  int step, maxStep;
  int size_x, size_y, me, x_domains,y_domains;
  int iconf[5];
  int size_x_glo, size_y_glo;
  double conf[2];

  /* MPI variables */
  int nproc, ndims;
  MPI_Comm comm, comm2d;
  int dims[2];
  int periods[2];
  int reorganisation = 0;
  MPI_Datatype column_type;
  int S=0, E=1, N=2, W=3;
  int NeighBor[4];
  int xcell, ycell, size_tot_x, size_tot_y;
  int *xs,*ys,*xe,*ye;
  double temp1_init, temp2_init, k0;
  FILE* file1;

  /* temp1_init: temperature init on edges */
  temp1_init = 10.0;

  /* temp2_init: temperature init inside */
  temp2_init = -10.0;

  /* diffusivity coefficient */
  k0 = 1;

  MPI_Init(&argc, &argv);
  comm = MPI_COMM_WORLD;
  MPI_Comm_size(comm,&nproc);
  MPI_Comm_rank(comm,&me);

  if (me==0) {
    readparam(iconf, conf);
  }
  MPI_Bcast(iconf,5,MPI_INT,0,comm);
  MPI_Bcast(conf,2,MPI_DOUBLE,0,comm);

  size_x    = iconf[0];
  size_y    = iconf[1];
  x_domains = iconf[2];
  y_domains = iconf[3];
  maxStep   = iconf[4];
  dt1        = conf[0];
  epsilon   = conf[1];

  if((me==0) && (nproc!=(x_domains*y_domains))) {
    printf("Number of processes not equal to Number of subdomains\n");
  }

  size_x_glo = size_x+2;
  size_y_glo = size_y+2;
  hx = 1.0/(double)(size_x_glo);
  hy = 1.0/(double)(size_y_glo);
  dt2     = 0.25*((min(hx,hy))*(min(hx,hy)))/k0;

  /* taking a good step for convergence */

  if(dt1>=dt2)
  { if(me==0)
    { printf("\n");
      printf("  Time step too large in 'param' file - Taking convergence criterion\n");}
      dt=dt2; }
  else { dt=dt1; }

  size_tot_x=size_x+2*x_domains+2;
  size_tot_y=size_y+2*y_domains+2;

  /* Allocation of 2D arrays */

  xfinal = malloc(size_x*size_y*sizeof(double));

  x0 = malloc(size_tot_y*sizeof(double*));
  x0_vals = malloc(size_tot_x*size_tot_y*sizeof(double));
  x = malloc(size_tot_y*sizeof(double*));
  x_vals = malloc(size_tot_x*size_tot_y*sizeof(double));

  for(j=0;j<=size_tot_y-1;j++) {
    x0[j] = &x0_vals[j*size_tot_x];
    x[j] = &x_vals[j*size_tot_x];
  }

  xs = malloc(nproc*sizeof(int));
  xe = malloc(nproc*sizeof(int));
  ys = malloc(nproc*sizeof(int));
  ye = malloc(nproc*sizeof(int));

  /* Create 2D cartesian grid */

  periods[0] = 0;
  periods[1] = 0;

  ndims = 2;
  dims[0]=x_domains;
  dims[1]=y_domains;

  MPI_Cart_create(comm, ndims, dims, periods, reorganisation, &comm2d);

  /* Identify neighbors */

  NeighBor[0] = MPI_PROC_NULL;
  NeighBor[1] = MPI_PROC_NULL;
  NeighBor[2] = MPI_PROC_NULL;
  NeighBor[3] = MPI_PROC_NULL;

  /* Left/West and right/Est neigbors */

  MPI_Cart_shift(comm2d,0,1,&NeighBor[W],&NeighBor[E]);

  /* Bottom/South and Upper/North neigbors */

  MPI_Cart_shift(comm2d,1,1,&NeighBor[S],&NeighBor[N]);

  /* coordinates of current cell */

  xcell=(size_x/x_domains);
  ycell=(size_y/y_domains);

  xtemp = malloc(xcell*ycell*sizeof(double));

  ys[me]=(y_domains-me%(y_domains)-1)*(ycell+2)+2;
  ye[me]=ys[me]+ycell-1;

  for(i=0;i<=y_domains-1;i++) 
  {xs[i]=2;}

  for(i=0;i<=y_domains-1;i++) 
  {xe[i]=xs[i]+xcell-1;}

  for(i=1;i<=(x_domains-1);i++)
  { for(j=0;j<=(y_domains-1);j++) 
    {
      xs[i*y_domains+j]=xs[(i-1)*y_domains+j]+xcell+2;
      xe[i*y_domains+j]=xs[i*y_domains+j]+xcell-1;
    }
  }

  /*  Create column data type to communicate with East and West neighbors */

  MPI_Type_vector( ycell, 1, size_tot_x, MPI_DOUBLE, &column_type); 
  MPI_Type_commit(&column_type);

  /* initialization */

  initvalues(x0, size_tot_x, size_tot_y, temp1_init, temp2_init);

  /* Update the boundaries */

  updateBound(x0, NeighBor, comm2d, column_type, me, xs, ys, xe, ye, xcell);

  step = 0;
  t    = 0.0;

  tstart = MPI_Wtime();

  /* UNTIL "Convergence" */

  while(!convergence)
  {  
    step = step + 1;
    t = t + dt ;

    /* perform one step of the explicit scheme */

    Explicit(x0, x, dt, &resLoc, hx, hy, me, xs, ys, xe, ye, k0);

    /* update the partial solution along the interface */

    updateBound(x0, NeighBor, comm2d, column_type, me, xs, ys, xe, ye, xcell);

    MPI_Allreduce(&resLoc, &result, 1, MPI_DOUBLE, MPI_SUM, comm);

    result= sqrt(result);

    if ((result<epsilon) || (step>maxStep)) break;
  }

  j=1;
  for(i=ys[me];i<=ye[me];i++)
  { for(k=0;k<=xcell-1;k++) 
    { xtemp[(j-1)*xcell+k]=x0[i][xs[me]+k];}
    j=j+1;
  }

  MPI_Gather(xtemp,xcell*ycell, MPI_DOUBLE, xfinal, xcell*ycell, MPI_DOUBLE, 0, comm);

  tend = MPI_Wtime();
  elapsed_time = tend-tstart;

  if (me == 0) {
    printf("\n");
    printf("  Time step = %3.18f\n", dt);
    printf("\n");
    printf("  Convergence = %11.9f after %d steps\n", epsilon, step);
    printf("\n");
    printf("  Problem size = %d\n", size_x*size_y);
    printf("\n");
    printf("  Wall Clock = %15.6f\n", elapsed_time);

    /* Print the solution at each point of the grid */

    printf("\n");
    printf("  Computed solution in outputMPI.dat\n");
    printf("\n");

    file1=fopen("outputMPI.dat","w");

    for(i=1;i<=size_x+1;i++)
    { fprintf(file1,"%15.11f ",temp1_init);}
    fprintf(file1,"%15.11f\n",temp1_init);

    for(i=1;i<=y_domains;i++)
    { for(j=1;j<=ycell;j++)
      { fprintf(file1,"%15.11f ",temp1_init);
        for(k=0;k<=x_domains-1;k++)
        { for(l=0;l<=xcell-1;l++)
          { fprintf(file1,"%15.11f ",xfinal[(j-1)*xcell+l+((y_domains-i)+k*y_domains)*xcell*ycell]);}
        }
        fprintf(file1,"%15.11f\n",temp1_init);
      }
    }

    for(i=1;i<=size_x+1;i++)
    { fprintf(file1,"%15.11f ",temp1_init);}
    fprintf(file1,"%15.11f\n",temp1_init);

    fclose(file1);
  }       

  free(x0);
  free(xfinal);
  free(x0_vals);
  free(xs);
  free(xe);
  free(ys);
  free(ye); 

  MPI_Type_free(&column_type);

  MPI_Finalize();

  return 0;
}        
